"""Computing CER/WER."""

import numpy as np
import pytest
from bio_parser import GLOBAL_STAT_NAME
from ie_eval.metrics.bag_of_entities import (
    compute_bag_of_entities,
    compute_bag_of_tagged_words,
    compute_bag_of_words,
)
from ie_eval.scorer import (
    BagOfWords,
    MicroAverageErrorRate,
    MicroAverageFScore,
)
from ie_eval.table_formatter import make_bag_of_entities_prettytable
from prettytable import PrettyTable

from tests.oi import FIXTURES


@pytest.mark.parametrize(
    (
        "labels",
        "predictions",
        "expected_all_words",
        "expected_label_vector",
        "expected_prediction_vector",
        "expected_insertions_deletions",
        "expected_substitutions",
        "expected_errors",
    ),
    [
        (
            ["Georges", "Washington", "cat"],
            ["Georg", "Washington", "cat"],
            ["Georg", "Georges", "Washington", "cat"],
            np.array([0, 1, 1, 1], dtype=int),
            np.array([1, 0, 1, 1], dtype=int),
            0,
            1,
            1,
        ),
        (
            ["Georges", "Washington", "cat"],
            ["Georges", "Washington"],
            ["Georges", "Washington", "cat"],
            np.array([1, 1, 1], dtype=int),
            np.array([1, 1, 0], dtype=int),
            1,
            0,
            1,
        ),
        (
            ["Georges", "Washington", "cat"],
            ["Georg", "Washington"],
            ["Georg", "Georges", "Washington", "cat"],
            np.array([0, 1, 1, 1], dtype=int),
            np.array([1, 0, 1, 0], dtype=int),
            1,
            1,
            2,
        ),
    ],
)
def test_bag_of_words_eval(
    labels,
    predictions,
    expected_all_words,
    expected_label_vector,
    expected_prediction_vector,
    expected_insertions_deletions,
    expected_substitutions,
    expected_errors,
):
    score = BagOfWords(labels, predictions)
    assert score.all_words == expected_all_words
    np.testing.assert_equal(score.label_word_vector, expected_label_vector)
    np.testing.assert_equal(score.prediction_word_vector, expected_prediction_vector)
    assert score.insertions_deletions == expected_insertions_deletions
    assert score.substitutions == expected_substitutions
    assert score.errors == expected_errors


@pytest.mark.parametrize(
    (
        "labels",
        "predictions",
        "expected_all_words",
        "expected_label_vector",
        "expected_prediction_vector",
        "expected_insertions_deletions",
        "expected_substitutions",
        "expected_errors",
    ),
    [
        (
            [("person", "Georges"), ("person", "Washington"), ("animal", "cat")],
            [("person", "Georg"), ("person", "Washington"), ("animal", "cat")],
            [
                ("animal", "cat"),
                ("person", "Georg"),
                ("person", "Georges"),
                ("person", "Washington"),
            ],
            np.array([1, 0, 1, 1], dtype=int),
            np.array([1, 1, 0, 1], dtype=int),
            0,
            1,
            1,
        ),
        (
            [("person", "Georges"), ("person", "Washington"), ("animal", "cat")],
            [("person", "Georges"), ("person", "Washington")],
            [("animal", "cat"), ("person", "Georges"), ("person", "Washington")],
            np.array([1, 1, 1], dtype=int),
            np.array([0, 1, 1], dtype=int),
            1,
            0,
            1,
        ),
        (
            [("person", "Georges"), ("person", "Washington"), ("animal", "cat")],
            [("person", "Georg"), ("animal", "Washington")],
            [
                ("animal", "Washington"),
                ("animal", "cat"),
                ("person", "Georg"),
                ("person", "Georges"),
                ("person", "Washington"),
            ],
            np.array([0, 1, 0, 1, 1], dtype=int),
            np.array([1, 0, 1, 0, 0], dtype=int),
            1,
            2,
            3,
        ),
    ],
)
def test_bag_of_tagged_words_eval(
    labels,
    predictions,
    expected_all_words,
    expected_label_vector,
    expected_prediction_vector,
    expected_insertions_deletions,
    expected_substitutions,
    expected_errors,
):
    score = BagOfWords(labels, predictions)
    assert score.all_words == expected_all_words
    np.testing.assert_equal(score.label_word_vector, expected_label_vector)
    np.testing.assert_equal(score.prediction_word_vector, expected_prediction_vector)
    assert score.insertions_deletions == expected_insertions_deletions
    assert score.substitutions == expected_substitutions
    assert score.errors == expected_errors


@pytest.mark.parametrize(
    (
        "labels",
        "predictions",
        "expected_all_words",
        "expected_label_vector",
        "expected_prediction_vector",
        "expected_insertions_deletions",
        "expected_substitutions",
        "expected_errors",
    ),
    [
        (
            [("person", "Georges Washington"), ("animal", "a fish called Wanda")],
            [("person", "Georges Washington"), ("animal", "a fish called Wanda")],
            [("animal", "a fish called Wanda"), ("person", "Georges Washington")],
            np.array([1, 1], dtype=int),
            np.array([1, 1], dtype=int),
            0,
            0,
            0,
        ),
        (
            [("person", "Georges Washington"), ("animal", "a fish called Wanda")],
            [("person", "Georges Washingtonn"), ("animal", "a fish called Wanda")],
            [
                ("animal", "a fish called Wanda"),
                ("person", "Georges Washington"),
                ("person", "Georges Washingtonn"),
            ],
            np.array([1, 1, 0], dtype=int),
            np.array([1, 0, 1], dtype=int),
            0,
            1,
            1,
        ),
        (
            [("person", "Georges Washington"), ("animal", "a fish called Wanda")],
            [("object", "Georges Washington"), ("animal", "a fish called Wanda")],
            [
                ("animal", "a fish called Wanda"),
                ("object", "Georges Washington"),
                ("person", "Georges Washington"),
            ],
            np.array([1, 0, 1], dtype=int),
            np.array([1, 1, 0], dtype=int),
            0,
            1,
            1,
        ),
        (
            [
                ("person", "Georges Washington"),
                ("person", "Hari Seldon"),
                ("animal", "a fish called Wanda"),
            ],
            [("person", "Hari Seldon"), ("animal", "a fish called Wanda")],
            [
                ("animal", "a fish called Wanda"),
                ("person", "Georges Washington"),
                ("person", "Hari Seldon"),
            ],
            np.array([1, 1, 1], dtype=int),
            np.array([1, 0, 1], dtype=int),
            1,
            0,
            1,
        ),
        (
            [
                ("person", "Georges Washington"),
                ("person", "Hari Seldon"),
                ("animal", "a fish called Wanda"),
            ],
            [("person", "Hari Seldon"), ("animal", "fish called Wanda")],
            [
                ("animal", "a fish called Wanda"),
                ("animal", "fish called Wanda"),
                ("person", "Georges Washington"),
                ("person", "Hari Seldon"),
            ],
            np.array([1, 0, 1, 1], dtype=int),
            np.array([0, 1, 0, 1], dtype=int),
            1,
            1,
            2,
        ),
    ],
)
def test_bag_of_entities_eval(
    labels,
    predictions,
    expected_all_words,
    expected_label_vector,
    expected_prediction_vector,
    expected_insertions_deletions,
    expected_substitutions,
    expected_errors,
):
    score = BagOfWords(labels, predictions)
    assert score.all_words == expected_all_words
    np.testing.assert_equal(score.label_word_vector, expected_label_vector)
    np.testing.assert_equal(score.prediction_word_vector, expected_prediction_vector)
    assert score.insertions_deletions == expected_insertions_deletions
    assert score.substitutions == expected_substitutions
    assert score.errors == expected_errors


@pytest.mark.parametrize(
    (
        "error_count",
        "true_positives",
        "false_positives",
        "false_negatives",
        "label_count",
        "count",
        "expected_table",
    ),
    [
        (
            {"per": 2, "date": 0, GLOBAL_STAT_NAME: 2},
            {"per": 10, "date": 20, GLOBAL_STAT_NAME: 30},
            {"per": 1, "date": 0, GLOBAL_STAT_NAME: 1},
            {"per": 1, "date": 0, GLOBAL_STAT_NAME: 1},
            {"per": 12, "date": 20, GLOBAL_STAT_NAME: 22},
            {"per": 5, "date": 5, GLOBAL_STAT_NAME: 5},
            """| Category | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:---------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| date     |   0.00   |     100.00    |   100.00   | 100.00 |      20 |           5 |
| per      |  16.67   |     90.91     |   90.91    | 90.91  |      12 |           5 |
| total    |   9.09   |     96.77     |   96.77    | 96.77  |      22 |           5 |""",
        ),
    ],
)
def test_make_pretty_table(
    error_count,
    true_positives,
    false_positives,
    false_negatives,
    label_count,
    count,
    expected_table,
) -> None:
    err_scores = MicroAverageErrorRate()
    err_scores.error_count = error_count
    err_scores.label_word_count = label_count
    err_scores.count = count
    det_scores = MicroAverageFScore()
    det_scores.true_positives = true_positives
    det_scores.false_positives = false_positives
    det_scores.false_negatives = false_negatives
    det_scores.label_word_count = label_count
    det_scores.count = count
    table = make_bag_of_entities_prettytable(detections=det_scores, errors=err_scores)
    assert isinstance(table, PrettyTable)
    assert table.field_names == [
        "Category",
        "bWER (%)",
        "Precision (%)",
        "Recall (%)",
        "F1 (%)",
        "N words",
        "N documents",
    ]
    assert table.get_string(sortby="Category") == expected_table


@pytest.mark.parametrize(
    (
        "label_dir",
        "prediction_dir",
        "shuffled_prediction_dir",
        "by_category",
        "expected_table",
    ),
    [
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            True,
            """| Category                | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:------------------------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| birth_date              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| cote_article            |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| cote_serie              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| fac                     |  80.00   |     100.00    |   20.00    | 33.33  |       5 |           1 |
| firstname               |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| gpe                     |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| husband_location        |   0.00   |     100.00    |   100.00   | 100.00 |       4 |           1 |
| husband_name            |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| husband_occupation      |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| husband_surname         |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| intitule                |  25.00   |     75.00     |   75.00    | 75.00  |      16 |           1 |
| link                    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| location_of_birth       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| norp                    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| ordinal                 |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| per                     |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| precisions_sur_cote     |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| surname                 |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| time                    |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| total                   |  25.45   |     81.48     |   80.00    | 80.73  |      55 |           5 |
| wife_name               |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wife_state              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_location   |   0.00   |     100.00    |   100.00   | 100.00 |       5 |           1 |
| wifes_father_name       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_occupation |   0.00   |     100.00    |   100.00   | 100.00 |       3 |           1 |
| wifes_father_surname    |   0.00   |     100.00    |   100.00   | 100.00 |       2 |           1 |
| wifes_mother_name       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |""",
        ),
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            False,
            """| Category | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:---------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| total    |  25.45   |     81.48     |   80.00    | 80.73  |      55 |           5 |""",
        ),
    ],
)
def test_compute_bag_of_words(
    label_dir,
    prediction_dir,
    shuffled_prediction_dir,
    by_category,
    expected_table,
):
    table = compute_bag_of_words(label_dir, prediction_dir, by_category)
    table_shuffled = compute_bag_of_words(
        label_dir,
        shuffled_prediction_dir,
        by_category,
    )
    assert table.get_string(sortby="Category") == expected_table
    assert table_shuffled.get_string(sortby="Category") == expected_table


@pytest.mark.parametrize(
    (
        "label_dir",
        "prediction_dir",
        "shuffled_prediction_dir",
        "by_category",
        "expected_table",
    ),
    [
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            True,
            """| Category                | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:------------------------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| birth_date              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| cote_article            |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| cote_serie              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| fac                     |  80.00   |     100.00    |   20.00    | 33.33  |       5 |           1 |
| firstname               |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| gpe                     |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| husband_location        |   0.00   |     100.00    |   100.00   | 100.00 |       4 |           1 |
| husband_name            |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| husband_occupation      |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| husband_surname         |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| intitule                |  25.00   |     75.00     |   75.00    | 75.00  |      16 |           1 |
| link                    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| location_of_birth       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| norp                    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| ordinal                 |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| per                     |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| precisions_sur_cote     |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| surname                 |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| time                    |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| total                   |  29.09   |     77.78     |   76.36    | 77.06  |      55 |           5 |
| wife_name               |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wife_state              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_location   |   0.00   |     100.00    |   100.00   | 100.00 |       5 |           1 |
| wifes_father_name       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_occupation |   0.00   |     100.00    |   100.00   | 100.00 |       3 |           1 |
| wifes_father_surname    |   0.00   |     100.00    |   100.00   | 100.00 |       2 |           1 |
| wifes_mother_name       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |""",
        ),
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            False,
            """| Category | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:---------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| total    |  29.09   |     77.78     |   76.36    | 77.06  |      55 |           5 |""",
        ),
    ],
)
def test_compute_bag_of_tagged_words(
    label_dir,
    prediction_dir,
    shuffled_prediction_dir,
    by_category,
    expected_table,
):
    table = compute_bag_of_tagged_words(label_dir, prediction_dir, by_category)
    table_shuffled = compute_bag_of_tagged_words(
        label_dir,
        shuffled_prediction_dir,
        by_category,
    )
    assert table.get_string(sortby="Category") == expected_table
    assert table_shuffled.get_string(sortby="Category") == expected_table


@pytest.mark.parametrize(
    (
        "label_dir",
        "prediction_dir",
        "shuffled_prediction_dir",
        "by_category",
        "expected_table",
    ),
    [
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            True,
            """| Category                | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:------------------------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| birth_date              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| cote_article            |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| cote_serie              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| fac                     |  100.00  |      0.00     |    0.00    |  0.00  |       2 |           1 |
| firstname               |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| gpe                     |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| husband_location        |   0.00   |     100.00    |   100.00   | 100.00 |       2 |           1 |
| husband_name            |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| husband_occupation      |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| husband_surname         |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| intitule                |  100.00  |      0.00     |    0.00    |  0.00  |       1 |           1 |
| link                    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| location_of_birth       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| norp                    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| ordinal                 |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| per                     |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| precisions_sur_cote     |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| surname                 |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| time                    |  100.00  |     100.00    |    0.00    |  0.00  |       1 |           1 |
| total                   |  32.14   |     71.43     |   71.43    | 71.43  |      28 |           5 |
| wife_name               |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wife_state              |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_location   |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_name       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_occupation |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_father_surname    |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |
| wifes_mother_name       |   0.00   |     100.00    |   100.00   | 100.00 |       1 |           1 |""",
        ),
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            False,
            """| Category | bWER (%) | Precision (%) | Recall (%) | F1 (%) | N words | N documents |
|:---------|:--------:|:-------------:|:----------:|:------:|--------:|------------:|
| total    |  32.14   |     71.43     |   71.43    | 71.43  |      28 |           5 |""",
        ),
    ],
)
def test_compute_bag_of_entities(
    label_dir,
    prediction_dir,
    shuffled_prediction_dir,
    by_category,
    expected_table,
):
    table = compute_bag_of_entities(label_dir, prediction_dir, by_category)
    table_shuffled = compute_bag_of_entities(
        label_dir,
        shuffled_prediction_dir,
        by_category,
    )
    assert table.get_string(sortby="Category") == expected_table
    assert table_shuffled.get_string(sortby="Category") == expected_table



from pathlib import Path

from bio_parser.utils import load_dataset
from prettytable import PrettyTable

from ie_eval.metrics.assignment_based import (
    compute_ecerewer,
    compute_nerval,
)
from ie_eval.metrics.bag_of_entities import WordType, compute_bag_of_anything
from ie_eval.table_formatter import make_summary_table


def compute_all_metrics(
    label_dir: Path,
    prediction_dir: Path,
    nerval_threshold: float = 30.0,
    by_category: bool = False,
) -> PrettyTable:
    """Compute all metrics.

    Args:
        label_dir (Path): Path to the label directory.
        prediction_dir (Path): Path to the prediction directory.
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False.

    Returns:
        PrettyTable: Summary table.
    """
    dataset = load_dataset(label_dir=label_dir, prediction_dir=prediction_dir)
    bow_table = compute_bag_of_anything(
        dataset,
        by_category,
        WordType.word,
        print_table=False,
    )
    botw_table = compute_bag_of_anything(
        dataset,
        by_category,
        WordType.tagged_word,
        print_table=False,
    )
    boe_table = compute_bag_of_anything(
        dataset,
        by_category,
        WordType.entity,
        print_table=False,
    )
    kwargs = {
        "label_dir": label_dir,
        "prediction_dir": prediction_dir,
        "by_category": by_category,
        "print_table": False,
    }
    oi_ecer_ewer_table = compute_ecerewer(order_independent=True, **kwargs)
    ecer_ewer_table = compute_ecerewer(order_independent=False, **kwargs)

    nerval_oi_table = compute_nerval(
        nerval_threshold=nerval_threshold,
        order_independent=True,
        **kwargs,
    )
    nerval_table = compute_nerval(
        nerval_threshold=nerval_threshold,
        order_independent=False,
        **kwargs,
    )

    table = make_summary_table(
        bow_table=bow_table,
        botw_table=botw_table,
        boe_table=boe_table,
        oi_ecer_ewer_table=oi_ecer_ewer_table,
        ecer_ewer_table=ecer_ewer_table,
        nerval_oi_table=nerval_oi_table,
        nerval_table=nerval_table,
    )
    print(table)  # noqa: T201
    return table



"""Compute CER and WER from a label/prediction dataset."""

import logging
from collections import defaultdict
from pathlib import Path
from typing import NamedTuple

import editdistance
from bio_parser import GLOBAL_STAT_NAME
from bio_parser.utils import load_dataset
from prettytable import MARKDOWN, PrettyTable

from ie_eval.metrics.utils import sort_categories
from ie_eval.scorer import (
    MicroAverageErrorRate,
    OiEcerEwer,
)
from ie_eval.table_formatter import (
    make_oi_ecer_ewer_prettytable,
)

logger = logging.getLogger(__name__)


def oiecerewer(
    label_dir: Path,
    prediction_dir: Path,
    by_category: bool = False,
    **kwargs,  # noqa: ARG001
) -> PrettyTable:
    """Compute reading-order-independent ECER and EWER metrics.

    Args:
        label_dir (Path): Path to the directory containing BIO label files.
        prediction_dir (Path): Path to the directory containing BIO prediction files.
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False. (Not implemented yet)
        print_table (bool, optional): Whether to print the table. Defaults to True.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    dataset = load_dataset(label_dir, prediction_dir)
    # Initialize scores
    ecer_total_score_err = MicroAverageErrorRate()
    ewer_total_score_err = MicroAverageErrorRate()

    # Iterate over the dataset
    for label, prediction in dataset:
        ecer_score = OiEcerEwer(
            labels=label.entities,
            predictions=prediction.entities,
            compute_ecer=True,
        )

        ewer_score = OiEcerEwer(
            labels=label.entities,
            predictions=prediction.entities,
            compute_ecer=False,
        )
        # Micro average
        ecer_total_score_err.update(GLOBAL_STAT_NAME, ecer_score)
        ewer_total_score_err.update(GLOBAL_STAT_NAME, ewer_score)

        # TODO: https://gitlab.teklia.com/ner/metrics/ie-eval/-/issues/12
        if by_category:
            continue

    # Format and display results
    table = make_oi_ecer_ewer_prettytable(
        ecer_total_score_err,
        ewer_total_score_err,
    )
    return table


class TextEval(NamedTuple):
    """Compute text errors between a label and prediction."""

    label: str
    """Label text."""
    prediction: str
    """Predicted text."""

    @property
    def char_errors(self) -> int:
        """Compute character errors between the label and prediction.

        Returns:
            Character errors.

        Examples:
            >>> TextEval("I really like cats", "I love cats").char_errors
            9
        """
        return editdistance.eval(
            format_string_for_cer(self.label),
            format_string_for_cer(self.prediction),
        )

    @property
    def word_errors(self) -> int:
        """Compute word errors between the label and prediction.

        Returns:
            Word errors.

        Examples:
            >>> TextEval("I really like cats", "I love cats").word_errors
            2
        """
        return editdistance.eval(
            format_string_for_wer(self.label),
            format_string_for_wer(self.prediction),
        )

    @property
    def char_totals(self) -> int:
        """Compute the max number of characters in the label or prediction.

        Returns:
            Number of characters.

        Examples:
            >>> TextEval("I really like cats", "I love cats").char_totals
            18
        """
        return max(
            len(format_string_for_cer(self.label)),
            len(format_string_for_cer(self.prediction)),
        )

    @property
    def word_totals(self) -> int:
        """Compute the max number of words in the label or prediction.

        Returns:
            Number of words.

        Examples:
            >>> TextEval("I really like cats", "I love cats").word_totals
            4
        """
        return max(
            len(format_string_for_wer(self.label)),
            len(format_string_for_wer(self.prediction)),
        )


class TotalScore:
    """Compute total evaluation scores."""

    def __init__(self):
        """Initialize errors and counts.

        Examples:
            >>> score = TotalScore()
        """
        self.char_errors = defaultdict(int)
        self.word_errors = defaultdict(int)
        self.char_totals = defaultdict(int)
        self.word_totals = defaultdict(int)
        self.count = defaultdict(int)

    def update(self, key, score: TextEval):
        """Update the score with the current evaluation for a given key.

        Args:
            key (str): Category to update.
            score (TextEval): Current score.

        Examples:
            >>> score.update("total", TextEval("I really like cats", "I like cats"))
            >>> score.update("animal", TextEval("cats", "cats"))
            >>> score.char_errors
            defaultdict(<class 'int'>, {'total': 7, 'animal': 0})
            >>> score.word_errors
            defaultdict(<class 'int'>, {'total': 1, 'animal': 0})
            >>> score.char_totals
            defaultdict(<class 'int'>, {'total': 18, 'animal': 4})
            >>> score.word_totals
            defaultdict(<class 'int'>, {'total': 4, 'animal': 1})
            >>> score.count
            defaultdict(<class 'int'>, {'total': 1, 'animal': 1})
        """
        self.char_errors[key] += score.char_errors
        self.word_errors[key] += score.word_errors
        self.char_totals[key] += score.char_totals
        self.word_totals[key] += score.word_totals
        self.count[key] += 1

    @property
    def categories(self) -> list[str]:
        """List of semantic categories for which scores are computed.

        Returns:
            The list of categories.

        Examples:
            >>> score.categories
            ['total', 'animal']
        """
        return list(self.count.keys())

    @property
    def cer(self) -> dict[str, float]:
        """Compute the Character Error Rate (%).

        Returns:
            The Character Error Rate.

        Examples:
            >>> score.cer
            {'total': 38.9, 'animal': 0.0}
        """
        return {
            key: round(100 * self.char_errors[key] / self.char_totals[key], 2)
            for key in self.categories
        }

    @property
    def wer(self) -> dict[str, float]:
        """Compute the Word Error Rate (%).

        Returns:
            The Word Error Rate.

        Examples:
            >>> score.wer
            {'total': 25.0, 'animal': 0.0}
        """
        return {
            key: round(100 * self.word_errors[key] / self.word_totals[key], 2)
            for key in self.categories
        }


def format_string_for_wer(text: str) -> list[str]:
    """Format string for WER computation.

    Args:
        text (str): The text to format.

    Returns:
        A list of words formatted for WER computation.

    Examples:
        >>> format_string_for_wer(text="this is a string to evaluate")
        ['this', 'is', 'a', 'string', 'to', 'evaluate']
        >>> format_string_for_wer(text="this is    another string to   evaluate")
        ['this', 'is', 'another', 'string', 'to', 'evaluate']
    """
    return text.strip().split()


def format_string_for_cer(text: str) -> str:
    """Format string for CER computation.

    Args:
        text (str): The text to format.

    Returns:
        The formatted text for CER computation.

    Examples:
        >>> format_string_for_cer(text="this is a string to evaluate")
        'this is a string to evaluate'
        >>> format_string_for_cer(text="this is    another string to   evaluate")
        'this is another string to evaluate'
    """
    return " ".join(text.strip().split())


def make_prettytable(score: TotalScore) -> PrettyTable:
    """Format and display results using PrettyTable.

    Args:
        score (TotalScore): Total scores.

    Returns:
        The evaluation table formatted in Markdown.
    """
    table = PrettyTable()
    table.set_style(MARKDOWN)
    table.field_names = ["Category", "ECER (%)", "EWER (%)", "Support"]
    table.align["Category"] = "l"
    table.align["Support"] = "r"

    rows = []
    for tag in score.categories:
        rows.append(
            [
                tag,
                f"{score.cer[tag]:.2f}",
                f"{score.wer[tag]:.2f}",
                score.count[tag],
            ],
        )

    table.add_rows(sort_categories(rows))
    return table


def merge_entities(entities: list[tuple[str, str]]) -> dict[str, str]:
    """Iterate over entities and merge text for each entity type.

    Args:
        entities (list[tuple[str, str]]): A list of entities.

    Returns:
        A dictionary with entity types as keys and the corresponding text as values.
    """
    entity_text = defaultdict(list)
    for tag, text in entities:
        entity_text[tag].append(text)
    return {k: " ".join(v) for k, v in entity_text.items()}


def cerwer(
    label_dir: Path,
    prediction_dir: Path,
    by_category: bool = False,
) -> PrettyTable:
    """Read BIO files and compute Character and Word Error Rates globally or for each NER category.

    Args:
        label_dir (Path): Path to the reference BIO file.
        prediction_dir (Path): Path to the prediction BIO file.
        by_category (bool): Whether to display CER/WER by category.

    Returns:
        A Markdown formatted table containing evaluation results.
    """
    # Initialize scores
    score = TotalScore()
    # Load the dataset
    dataset = load_dataset(label_dir, prediction_dir)
    # Iterate over the dataset
    for label, prediction in dataset:
        # Compute global CER and WER
        score.update(GLOBAL_STAT_NAME, TextEval(label.text, prediction.text))

        # Compute CER and WER by category
        if not by_category:
            continue
        label_text = merge_entities(label.entities)
        pred_text = merge_entities(prediction.entities)
        for tag in label_text:
            score.update(
                tag,
                TextEval(
                    label_text[tag],
                    pred_text.get(tag, ""),
                ),
            )

    # Format and display results
    return make_prettytable(score)



"""Computing CER/WER."""

import pytest
from bio_parser import GLOBAL_STAT_NAME
from ie_eval.metrics.cer import (
    TextEval,
    TotalScore,
    cerwer,
    format_string_for_cer,
    format_string_for_wer,
    make_prettytable,
    merge_entities,
)
from prettytable import PrettyTable

from tests.od import FIXTURES


@pytest.mark.parametrize(
    ("text", "expected_formatted_text"),
    [
        (
            "I took a train from Bordeaux to Nantes",
            "I took a train from Bordeaux to Nantes",
        ),
        (
            " I took a train from Bordeaux to Nantes ",
            "I took a train from Bordeaux to Nantes",
        ),
        (
            " I took a train from     Bordeaux to Nantes  ",
            "I took a train from Bordeaux to Nantes",
        ),
    ],
)
def test_format_string_for_cer(text, expected_formatted_text):
    assert format_string_for_cer(text) == expected_formatted_text


@pytest.mark.parametrize(
    ("text", "expected_formatted_text"),
    [
        (
            "I took a train from Bordeaux to Nantes",
            ["I", "took", "a", "train", "from", "Bordeaux", "to", "Nantes"],
        ),
        (
            " I took     a train from Bordeaux    to Nantes   ",
            ["I", "took", "a", "train", "from", "Bordeaux", "to", "Nantes"],
        ),
    ],
)
def test_format_string_for_wer(text, expected_formatted_text):
    assert format_string_for_wer(text) == expected_formatted_text


@pytest.mark.parametrize(
    (
        "label",
        "prediction",
        "expected_char_errors",
        "expected_word_errors",
        "expected_total_chars",
        "expected_total_words",
    ),
    [
        (
            "I took a train from Bordeaux to Nantes",
            " I took     a train from Bordeaux    to Nantes   ",
            0,
            0,
            38,
            8,
        ),
        (
            "I took a train from Bordeaux to Nantes",
            "I took the plane from Bordeaux to Nantes",
            7,
            2,
            40,
            8,
        ),
        (
            "I took a train from Bordeaux to Nantes",
            "I like trains",
            28,
            7,
            38,
            8,
        ),
    ],
)
def test_text_eval(
    label,
    prediction,
    expected_char_errors,
    expected_word_errors,
    expected_total_chars,
    expected_total_words,
):
    score = TextEval(label, prediction)
    assert score.char_errors == expected_char_errors
    assert score.word_errors == expected_word_errors
    assert score.char_totals == expected_total_chars
    assert score.word_totals == expected_total_words


@pytest.mark.parametrize(
    (
        "errors_chars",
        "errors_words",
        "total_chars",
        "total_words",
        "total_counts",
        "expected_table",
    ),
    [
        (
            {"per": 13, "date": 0, GLOBAL_STAT_NAME: 17},
            {"per": 2, "date": 0, GLOBAL_STAT_NAME: 14},
            {"per": 123, "date": 100, GLOBAL_STAT_NAME: 250},
            {"per": 10, "date": 19, GLOBAL_STAT_NAME: 78},
            {"per": 20, "date": 13, GLOBAL_STAT_NAME: 50},
            """| Category | ECER (%) | EWER (%) | Support |
|:---------|:--------:|:--------:|--------:|
| date     |   0.00   |   0.00   |      13 |
| per      |  10.57   |  20.00   |      20 |
| total    |   6.80   |  17.95   |      50 |""",
        ),
    ],
)
def test_make_pretty_table(
    errors_chars,
    errors_words,
    total_chars,
    total_words,
    total_counts,
    expected_table,
):
    scores = TotalScore()
    scores.char_errors = errors_chars
    scores.word_errors = errors_words
    scores.char_totals = total_chars
    scores.word_totals = total_words
    scores.count = total_counts
    table = make_prettytable(scores)
    assert isinstance(table, PrettyTable)
    assert table.field_names == [
        "Category",
        "ECER (%)",
        "EWER (%)",
        "Support",
    ]
    assert table.get_string() == expected_table


@pytest.mark.parametrize(
    ("entities", "expected_dict"),
    [
        (
            [("per", "Marie"), ("date", "December"), ("event", "Christmas")],
            {"per": "Marie", "date": "December", "event": "Christmas"},
        ),
        (
            [
                ("org", "Church"),
                ("per", "Christ"),
                ("event", "Lord's Supper"),
                ("per", "James"),
                ("work_of_art", "Bible"),
            ],
            {
                "org": "Church",
                "per": "Christ James",
                "event": "Lord's Supper",
                "work_of_art": "Bible",
            },
        ),
        (
            [("per", "Anne"), ("per", "Marie"), ("per", "Louise")],
            {"per": "Anne Marie Louise"},
        ),
    ],
)
def test_merge_entities(entities, expected_dict):
    assert merge_entities(entities) == expected_dict


@pytest.mark.parametrize(
    ("label_dir", "prediction_dir", "by_category", "expected_table"),
    [
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            True,
            """| Category    | ECER (%) | EWER (%) | Support |
|:------------|:--------:|:--------:|--------:|
| date        |   0.00   |   0.00   |       1 |
| event       |  100.00  |  100.00  |       1 |
| ordinal     |  100.00  |  100.00  |       1 |
| org         |  45.45   |  50.00   |       1 |
| per         |  32.14   |  33.33   |       2 |
| work_of_art |  100.00  |  100.00  |       1 |
| total       |   3.53   |  11.79   |       3 |""",
        ),
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            False,
            """| Category | ECER (%) | EWER (%) | Support |
|:---------|:--------:|:--------:|--------:|
| total    |   3.53   |  11.79   |       3 |""",
        ),
    ],
)
def test_compute_cer_wer(label_dir, prediction_dir, by_category, expected_table):
    table = cerwer(label_dir, prediction_dir, by_category)
    assert table.get_string() == expected_table



import pytest
from ie_eval.metrics.nerval import nerval

from tests.od import FIXTURES


@pytest.mark.parametrize("by_category", [True, False])
def test_compute_nerval(by_category):
    table = nerval(
        label_dir=FIXTURES / "labels",
        prediction_dir=FIXTURES / "predictions",
        nerval_threshold=30,
        by_category=by_category,
    )

    # Check the printed Markdown table
    suffix = "_by_category" if by_category else ""
    assert table.get_string() == (FIXTURES / f"nerval{suffix}.md").read_text()



"""Compute the bag-of-words/bag-of-tagged-words/bag-of-entities metrics from a label/prediction dataset."""

import logging
from collections import defaultdict
from enum import Enum
from pathlib import Path

from bio_parser import GLOBAL_STAT_NAME
from bio_parser.parse.document import Document
from bio_parser.utils import load_dataset
from prettytable import PrettyTable

from ie_eval.scorer import (
    BagOfWords,
    MicroAverageErrorRate,
    MicroAverageFScore,
)
from ie_eval.table_formatter import make_bag_of_entities_prettytable

logger = logging.getLogger(__name__)


class WordType(Enum):
    """Word Type."""

    word = "word"
    tagged_word = "tagged_word"
    entity = "entity"


def prepare(document: Document, attr_name: str, with_category: bool = False):
    """Get list of words, tagged words or entities (overall and by category).

    Args:
        document (Document): Processed document
        attr_name (str): Name of the attribute which holds the objects to store
        with_category (bool, optional): Store the category along the word. Defaults to False.
    """
    prepared_data = defaultdict(list)
    for category, text in getattr(document, attr_name):
        data = (category, text) if with_category else text
        prepared_data[category].append(data)
        prepared_data[GLOBAL_STAT_NAME].append(data)
    return prepared_data


def tokenize_entities(
    label: Document,
    prediction: Document,
    word_type: WordType,
) -> tuple[
    dict[str, list[str | tuple[str, str]]],
    dict[str, list[str | tuple[str, str]]],
]:
    """Prepare label and prediction for BagOfWord computation.

    Args:
        label (Document): the label document
        prediction (Document): the prediction document
        word_type (WordType): Whether to consider a list of words, list tagged words, or list of tagged entities.

    Returns:
        a label dictionary with categories as keys and corresponding list of words, tagged words or tagged entities as values.
        a prediction dictionary with categories as keys and corresponding list of words, tagged words or tagged entities as values.
    """
    kwargs = {}
    match word_type:
        case WordType.word:
            # Return list of words
            kwargs = {"attr_name": "word_entities"}
        case WordType.tagged_word:
            # Return list of tagged words
            kwargs = {"attr_name": "word_entities", "with_category": True}
        case WordType.entity:
            # Return list of tagged entities
            kwargs = {"attr_name": "entities", "with_category": True}
    return (prepare(document=label, **kwargs), prepare(document=prediction, **kwargs))


def compute_bag_of_anything(
    dataset: list[tuple[Document, Document]],
    by_category: bool = False,
    word_type: WordType = WordType.word,
    print_table: bool = True,
) -> PrettyTable:
    """Compute bag-of-words, bag-of-tagged-words, or bag-of-entities.

    Args:
        dataset (list[tuple[Document, Document]]): a dataset containing a list of tuple with the label and corresponding prediction.
        word_type (WordType): Type of words to use for bag-of-words computation.
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False.
        print_table (bool, optional): Whether to print the table. Defaults to True.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    # Initialize scores
    total_score_f1 = MicroAverageFScore()
    total_score_err = MicroAverageErrorRate()

    # Iterate over the dataset
    for label, prediction in dataset:
        # Compute scores
        categories = {entity for entity, _word in label.word_entities}
        label_list, prediction_list = tokenize_entities(
            label,
            prediction,
            word_type=word_type,
        )

        score = BagOfWords(
            labels=label_list[GLOBAL_STAT_NAME],
            predictions=prediction_list[GLOBAL_STAT_NAME],
        )

        # Micro average
        total_score_f1.update(GLOBAL_STAT_NAME, score)
        total_score_err.update(GLOBAL_STAT_NAME, score)

        # Compute bag-of-tagged words by category
        if not by_category:
            continue

        for category in categories:
            # Compute scores
            category_score = BagOfWords(
                label_list.get(category, []),
                prediction_list.get(category, []),
            )

            # Micro average
            total_score_err.update(category, category_score)
            total_score_f1.update(category, category_score)

    # Format and display results
    table = make_bag_of_entities_prettytable(
        errors=total_score_err,
        detections=total_score_f1,
    )
    if print_table:
        print(table)  # noqa: T201
    return table


def compute_bag_of_words(
    label_dir: Path,
    prediction_dir: Path,
    by_category: bool = False,
) -> PrettyTable:
    """Compute bag-of-words.

    Args:
        label_dir (Path): Path to the label directory.
        prediction_dir (Path): Path to the prediction directory.
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    return compute_bag_of_anything(
        load_dataset(label_dir=label_dir, prediction_dir=prediction_dir),
        by_category=by_category,
        word_type=WordType.word,
    )


def compute_bag_of_tagged_words(
    label_dir: Path,
    prediction_dir: Path,
    by_category: bool = False,
) -> PrettyTable:
    """Compute bag-of-tagged-words.

    Args:
        label_dir (Path): Path to the label directory.
        prediction_dir (Path): Path to the prediction directory.
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    return compute_bag_of_anything(
        load_dataset(label_dir=label_dir, prediction_dir=prediction_dir),
        by_category=by_category,
        word_type=WordType.tagged_word,
    )


def compute_bag_of_entities(
    label_dir: Path,
    prediction_dir: Path,
    by_category: bool = False,
) -> PrettyTable:
    """Compute bag-of-entities.

    Args:
        label_dir (Path): Path to the label directory.
        prediction_dir (Path): Path to the prediction directory.
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    return compute_bag_of_anything(
        load_dataset(label_dir=label_dir, prediction_dir=prediction_dir),
        by_category=by_category,
        word_type=WordType.entity,
    )



"""PrettyTable formatters."""

import json

from prettytable import MARKDOWN, PrettyTable

from ie_eval import GLOBAL_STAT_NAME
from ie_eval.scorer import MicroAverageErrorRate, MicroAverageFScore


def make_summary_table(
    bow_table: PrettyTable,
    botw_table: PrettyTable,
    boe_table: PrettyTable,
    oi_ecer_ewer_table: PrettyTable,
    ecer_ewer_table: PrettyTable,
    nerval_oi_table: PrettyTable,
    nerval_table: PrettyTable,
) -> PrettyTable:
    """Format and display a summary table from all available metrics.

    Args:
        bow_table (PrettyTable): Bag-of-word table.
        botw_table (PrettyTable): Bag-of-tagged-word table.
        boe_table (PrettyTable): Bag-of-entity table.
        oi_ecer_ewer_table (PrettyTable): ECER/EWER OI table.
        ecer_ewer_table (PrettyTable): ECER/EWER table.
        nerval_oi_table (PrettyTable): Nerval OI table.
        nerval_table (PrettyTable): Nerval table.

    Returns:
        The summary evaluation table formatted in Markdown.
    """
    summary_table = PrettyTable()
    summary_table.set_style(MARKDOWN)
    summary_table.field_names = [
        "Category",
        "BoW-F1 (%)",
        "BoTW-F1 (%)",
        "BoE-F1 (%)",
        "ECER (%)",
        "EWER (%)",
        "ECER OI (%)",
        "EWER OI (%)",
        "Nerval-F1 (OI) (%)",
        "Nerval-F1 (%)",
        "N documents",
    ]
    summary_table.align["Category"] = "l"
    summary_table.align["N documents"] = "r"
    for i in range(1, len(json.loads(bow_table.get_json_string()))):
        summary_table.add_row(
            [
                json.loads(bow_table.get_json_string())[i]["Category"],
                json.loads(bow_table.get_json_string())[i]["F1 (%)"],
                json.loads(botw_table.get_json_string())[i]["F1 (%)"],
                json.loads(boe_table.get_json_string())[i]["F1 (%)"],
                json.loads(oi_ecer_ewer_table.get_json_string())[i]["ECER (%)"],
                json.loads(oi_ecer_ewer_table.get_json_string())[i]["EWER (%)"],
                json.loads(ecer_ewer_table.get_json_string())[i]["ECER (%)"],
                json.loads(ecer_ewer_table.get_json_string())[i]["EWER (%)"],
                json.loads(nerval_oi_table.get_json_string())[i]["F1 (%)"],
                json.loads(nerval_table.get_json_string())[i]["F1 (%)"],
                json.loads(bow_table.get_json_string())[i]["N documents"],
            ],
        )
    return summary_table


def make_bag_of_entities_prettytable(
    errors: MicroAverageErrorRate,
    detections: MicroAverageFScore,
) -> PrettyTable:
    """Format and display Bag-of-Word results using PrettyTable.

    Args:
        errors (MicroAverageErrorRate): Total error rates (bWER).
        detections (MicroAverageFScore): Total recognition rates (Precision, Recall, F1).

    Returns:
        The evaluation table formatted in Markdown.
    """
    table = PrettyTable()
    table.set_style(MARKDOWN)
    table.field_names = [
        "Category",
        "bWER (%)",
        "Precision (%)",
        "Recall (%)",
        "F1 (%)",
        "N words",
        "N documents",
    ]
    table.align["Category"] = "l"
    table.align["N words"] = "r"
    table.align["N documents"] = "r"
    for tag in errors.categories:
        table.add_row(
            [
                tag,
                f"{errors.error_rate[tag]:.2f}",
                f"{detections.precision[tag]:.2f}",
                f"{detections.recall[tag]:.2f}",
                f"{detections.f1[tag]:.2f}",
                errors.label_word_count[tag],
                errors.count[tag],
            ],
        )
    return table


def make_oi_ecer_ewer_prettytable(
    ecer_total_score_err: MicroAverageErrorRate,
    ewer_total_score_err: MicroAverageErrorRate,
) -> PrettyTable:
    """Format and display order independent ECER/EWER results using PrettyTable."""
    table = PrettyTable()
    table.set_style(MARKDOWN)
    table.field_names = [
        "Category",
        "ECER (%)",
        "EWER (%)",
        "N entities",
        "N documents",
    ]
    table.align["Category"] = "l"
    table.align["N entities"] = "r"
    table.align["N documents"] = "r"
    for tag in ecer_total_score_err.categories:
        table.add_row(
            [
                tag,
                f"{ecer_total_score_err.error_rate[tag]:.2f}",
                f"{ewer_total_score_err.error_rate[tag]:.2f}",
                ecer_total_score_err.label_word_count[tag],
                ecer_total_score_err.count[tag],
            ],
        )
    return table


def make_oi_nerval_prettytable(
    detections: MicroAverageFScore,
) -> PrettyTable:
    """Format and display order independent Nerval results using PrettyTable."""
    table = PrettyTable()
    table.set_style(MARKDOWN)
    table.field_names = [
        "Category",
        "Precision (%)",
        "Recall (%)",
        "F1 (%)",
        "N entities",
        "N documents",
    ]
    table.align["Category"] = "l"
    table.align["N entities"] = "r"
    table.align["N documents"] = "r"
    for tag in detections.categories:
        table.add_row(
            [
                tag,
                f"{detections.precision[tag]:.2f}",
                f"{detections.recall[tag]:.2f}",
                f"{detections.f1[tag]:.2f}",
                detections.label_word_count[tag],
                detections.count[tag],
            ],
        )
    return table


def make_nerval_pretty_table(rows: list[list]) -> PrettyTable:
    """Format into a pretty table.

    Args:
        rows (list[list]): Rows of the table

    Returns:
        PrettyTable: Formatted table.
    """
    table = PrettyTable()
    table.field_names = ["Category", "Precision (%)", "Recall (%)", "F1 (%)", "Support"]
    table.set_style(MARKDOWN)
    # First column should be left aligned still
    table.align["Category"] = "l"

    def _special_sort(row: list[str]) -> str:
        if row[0] == GLOBAL_STAT_NAME:
            # Place the line for all entities at the very top
            return ""
        return row[0]

    rows.sort(key=_special_sort)
    # Place ALL_ENTITIES row at the end
    rows.append(rows.pop(0))

    table.add_rows(rows)
    return table



"""Compute Precision, Recall and F1 from a label/prediction dataset."""

from collections import Counter, defaultdict
from operator import itemgetter
from pathlib import Path

from bio_parser import GLOBAL_STAT_NAME
from bio_parser.parse.document import Document
from bio_parser.utils import load_dataset
from nerval.evaluate import evaluate
from prettytable import PrettyTable

from ie_eval.metrics.utils import sort_categories
from ie_eval.scorer import MicroAverageFScore, OiNerval
from ie_eval.table_formatter import make_nerval_pretty_table, make_oi_nerval_prettytable

METRICS = ["predicted", "matched", "Support"]
NERVAL_GLOBAL_STAT = "ALL"


def _format_document(document: Document) -> dict:
    return {
        # Support cases where the string is empty (edlib will crash.)
        "words": document.text or "∅",
        "labels": document.char_labels or ["O"],
        "entity_count": {
            NERVAL_GLOBAL_STAT: len(document.entities),
            **Counter(map(itemgetter(0), document.entities)),
        },
    }


def compute_precision(matched: int, predicted: int) -> float:
    """Compute precision."""
    if predicted == 0:
        return 100 if matched == 0 else 0
    return 100 * matched / predicted


def compute_recall(matched: int, support: int) -> float:
    """Compute recall."""
    if support == 0:
        return 100 if matched == 0 else 0
    return 100 * matched / support


def compute_f1(precision: float, recall: float) -> float:
    """Compute F1 score."""
    if precision + recall == 0:
        return 0
    return 2 * precision * recall / (precision + recall)


def nerval(
    label_dir: Path,
    prediction_dir: Path,
    nerval_threshold: float,
    by_category: bool = False,
) -> PrettyTable:
    """Read BIO files and compute Precision, Recall and F1 globally or for each NER category.

    Args:
        label_dir (Path): Path to the reference BIO file.
        prediction_dir (Path): Path to the prediction BIO file.
        nerval_threshold (float): Character Error Rate threshold used to match entities.
        by_category (bool): Whether to display Precision/Recall/F1 by category.

    Returns:
        A Markdown formatted table containing evaluation results.
    """
    # Load the dataset
    dataset = load_dataset(label_dir, prediction_dir)

    # Iterate over the dataset
    scores = defaultdict(lambda: defaultdict(int))
    for label, prediction in dataset:
        # if label.bio_repr == "":
        #     continue
        cor_score = evaluate(
            _format_document(label),
            _format_document(prediction),
            nerval_threshold,
        )
        for entity, results in cor_score.items():
            # Nerval uses a different global statistic than us
            if entity == NERVAL_GLOBAL_STAT:
                entity = GLOBAL_STAT_NAME

            for metric in METRICS:
                scores[entity][metric] += results[metric] or 0

    results = []
    for entity, score in scores.items():
        if entity != GLOBAL_STAT_NAME and not by_category:
            continue
        precision = compute_precision(score["matched"], score["predicted"])
        recall = compute_recall(score["matched"], score["Support"])
        results.append(
            [
                entity,
                round(precision, 2),
                round(recall, 2),
                round(compute_f1(precision, recall), 2),
                score["Support"],
            ],
        )

    return make_nerval_pretty_table(
        rows=sort_categories(results),
    )


def oinerval(
    label_dir: Path,
    prediction_dir: Path,
    nerval_threshold: float = 30.0,
    by_category: bool = False,
) -> PrettyTable:
    """Compute reading-order-independent Nerval Precision, Recall and F1 scores.

    Args:
        label_dir (Path): Path to the directory containing BIO label files.
        prediction_dir (Path): Path to the directory containing BIO prediction files.
        nerval_threshold (float, default = 30.0): Threshold for the amount of character error that is tolerable during the computation (values in the range [0.0, 100.0]])
        by_category (bool, optional): Whether to compute the metric globally or for each category. Defaults to False. (Not implemented yet)
        print_table (bool, optional): Whether to print the table. Defaults to True.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    dataset = load_dataset(label_dir, prediction_dir)
    # Initialize scores
    total_score_f1 = MicroAverageFScore()

    # Iterate over the dataset
    for label, prediction in dataset:
        score = OiNerval(
            labels=label.entities,
            predictions=prediction.entities,
            nerval_threshold=nerval_threshold,
        )

        # Micro average
        total_score_f1.update(GLOBAL_STAT_NAME, score)

        # TODO: https://gitlab.teklia.com/ner/metrics/ie-eval/-/issues/12
        if not by_category:
            continue

    # Format and display results
    table = make_oi_nerval_prettytable(
        total_score_f1,
    )
    return table



"""Utils."""

from ie_eval import GLOBAL_STAT_NAME


def sort_categories(categories: list[list[str]]) -> list[list[str]]:
    """Sort a list of categories with their associated metrics.

    All categories are alphabetically sorted except for GLOBAL_STAT_NAME
    which is appended at the very end of the sorted list.

    Args:
        categories (list[list[str]]): List of categories with their metrics.

    Returns:
        list[list[str]]: A sorted version of the provided list of categories.
    """
    sorted_categories = sorted(categories)
    return sorted(
        sorted_categories,
        key=lambda e: (
            sorted_categories.index(e)
            if e[0] != GLOBAL_STAT_NAME
            else len(sorted_categories)
        ),
    )



"""Command Line Arguments"""

import argparse
from pathlib import Path

from ie_eval.metrics import compute_all_metrics
from ie_eval.metrics.assignment_based import (
    compute_ecerewer,
    compute_nerval,
)
from ie_eval.metrics.bag_of_entities import (
    compute_bag_of_entities,
    compute_bag_of_tagged_words,
    compute_bag_of_words,
)


def add_bow_parser(commands):
    """Add parser for the bag-of-words computation subcommand."""
    parser = commands.add_parser(
        "bow",
        help="Compute bag-of-entities metrics.",
    )
    parser.set_defaults(func=compute_bag_of_words)
    parser.add_argument(
        "--label-dir",
        "-l",
        type=Path,
        required=True,
        help="Path to the directory containing BIO label files.",
    )
    parser.add_argument(
        "--prediction-dir",
        "-p",
        type=Path,
        required=True,
        help="Path to the directory containing BIO prediction files.",
    )
    parser.add_argument(
        "--by-category",
        "-c",
        action="store_true",
        help="Whether to compute scores for each semantic category.",
    )


def add_botw_parser(commands):
    """Add parser for the bag-of-tagged-words computation subcommand."""
    parser = commands.add_parser(
        "botw",
        help="Compute bag-of-tagged-words metrics.",
    )
    parser.set_defaults(func=compute_bag_of_tagged_words)
    parser.add_argument(
        "--label-dir",
        "-l",
        type=Path,
        required=True,
        help="Path to the directory containing BIO label files.",
    )
    parser.add_argument(
        "--prediction-dir",
        "-p",
        type=Path,
        required=True,
        help="Path to the directory containing BIO prediction files.",
    )
    parser.add_argument(
        "--by-category",
        "-c",
        action="store_true",
        help="Whether to compute scores for each semantic category.",
    )


def add_boe_parser(commands):
    """Add parser for the bag-of-entities computation subcommand."""
    parser = commands.add_parser(
        "boe",
        help="Compute bag-of-entities metrics.",
    )
    parser.set_defaults(func=compute_bag_of_entities)
    parser.add_argument(
        "--label-dir",
        "-l",
        type=Path,
        required=True,
        help="Path to the directory containing BIO label files.",
    )
    parser.add_argument(
        "--prediction-dir",
        "-p",
        type=Path,
        required=True,
        help="Path to the directory containing BIO prediction files.",
    )
    parser.add_argument(
        "--by-category",
        "-c",
        action="store_true",
        help="Whether to compute scores for each semantic category.",
    )


def add_ecer_ewer_parser(commands):
    """Add parser for the ECER and EWER computation subcommand."""
    parser = commands.add_parser(
        "cer",
        help="Compute ECER and EWER metrics.",
    )
    parser.set_defaults(func=compute_ecerewer)
    parser.add_argument(
        "--label-dir",
        "-l",
        type=Path,
        required=True,
        help="Path to the directory containing BIO label files.",
    )
    parser.add_argument(
        "--prediction-dir",
        "-p",
        type=Path,
        required=True,
        help="Path to the directory containing BIO prediction files.",
    )
    parser.add_argument(
        "--order-independent",
        "-oi",
        action="store_true",
        help="Make the metric order-independent.",
    )
    parser.add_argument(
        "--by-category",
        "-c",
        action="store_true",
        help="Whether to compute scores for each semantic category. Only available for the order-dependent version.",
    )


def add_nerval_parser(commands):
    """Add parser for the Nerval computation subcommand."""
    parser = commands.add_parser(
        "nerval",
        help="Compute Nerval Precision, Recall and F1 scores.",
    )
    parser.set_defaults(func=compute_nerval)
    parser.add_argument(
        "--label-dir",
        "-l",
        type=Path,
        required=True,
        help="Path to the directory containing BIO label files.",
    )
    parser.add_argument(
        "--prediction-dir",
        "-p",
        type=Path,
        required=True,
        help="Path to the directory containing BIO prediction files.",
    )
    parser.add_argument(
        "--nerval-threshold",
        "-t",
        type=float,
        default=0.0,
        help="Threshold of acceptable character errors, must be in the range: [0.0, 100.0]",
    )
    parser.add_argument(
        "--order-independent",
        "-oi",
        action="store_true",
        help="Make the metric order-independent.",
    )
    parser.add_argument(
        "--by-category",
        "-c",
        action="store_true",
        help="Whether to compute scores for each semantic category. Only available for the order-dependent version.",
    )


def add_summary_parser(commands):
    """Add parser subcommand to compute all metrics."""
    parser = commands.add_parser(
        "all",
        help="Compute all metrics.",
    )
    parser.set_defaults(func=compute_all_metrics)
    parser.add_argument(
        "--label-dir",
        "-l",
        type=Path,
        required=True,
        help="Path to the directory containing BIO label files.",
    )
    parser.add_argument(
        "--prediction-dir",
        "-p",
        type=Path,
        required=True,
        help="Path to the directory containing BIO prediction files.",
    )
    parser.add_argument(
        "--nerval-threshold",
        "-t",
        type=float,
        default=0.0,
        help="Threshold of acceptable character errors for Nerval, must be in the range: [0.0, 100.0]",
    )
    parser.add_argument(
        "--by-category",
        "-c",
        action="store_true",
        help="Whether to compute scores for each semantic category.",
    )


def main():
    parser = argparse.ArgumentParser(
        prog="ie-eval",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    commands = parser.add_subparsers(
        help="Available metrics to evaluate IE models.",
    )
    add_bow_parser(commands)
    add_botw_parser(commands)
    add_boe_parser(commands)
    add_ecer_ewer_parser(commands)
    add_nerval_parser(commands)
    add_summary_parser(commands)

    args = vars(parser.parse_args())
    if "func" in args:
        args.pop("func")(**args)
    else:
        parser.print_help()



"""Computing CER/WER."""

import math

import pytest
from ie_eval.metrics.cer import oiecerewer
from ie_eval.metrics.nerval import oinerval
from ie_eval.scorer import (
    OiEcerEwer,
    OiNerval,
    calc_dist_sus_entity,
)

from tests.oi import FIXTURES


@pytest.mark.parametrize(
    (
        "label",
        "prediction",
        "char_level",
        "expected_substitution_cost",
    ),
    [
        (
            ("Person", "George Washington"),
            ("Place", "George Washington"),
            True,
            1.0,
        ),
        (
            ("Person", "George Washington"),
            ("Place", "George Washington"),
            False,
            1.0,
        ),
        (
            ("Person", "George Washington"),
            ("Person", "George Washington"),
            True,
            0.0,
        ),
        (
            ("Person", "George Washington"),
            ("Person", "George Washington"),
            False,
            0.0,
        ),
        (
            ("Person", "George Washington"),
            ("Person", "George Washingtn"),
            True,
            0.058823529411764705,
        ),
        (
            ("Person", "George Washington"),
            ("Person", "George Washingtn"),
            False,
            0.5,
        ),
    ],
)
def test_calc_dist_sus_entity(
    label,
    prediction,
    char_level,
    expected_substitution_cost,
):
    sust_cost = calc_dist_sus_entity(prediction, label, char_level)
    assert math.isclose(expected_substitution_cost, sust_cost)


@pytest.mark.parametrize(
    (
        "labels",
        "predictions",
        "char_level",
        "expected_num_ne_gt",
        "expected_num_ne_hyp",
        "expected_cost_matrix",
        "expected_errors",
    ),
    [
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Person", "George Washingtn")],
            True,
            2,
            1,
            [[0.058823529411764705, 1.0], [1.0, 1.0]],
            1.058823529411764705,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Person", "George Washingtn")],
            False,
            2,
            1,
            [[0.5, 1.0], [1.0, 1.0]],
            1.5,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Place", "Valencia")],
            True,
            2,
            1,
            [[1.0, 1.0], [0.0, 1.0]],
            1.0,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Place", "Valencia")],
            False,
            2,
            1,
            [[1.0, 1.0], [0.0, 1.0]],
            1.0,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [],
            False,
            2,
            0,
            [[1.0, 1.0], [1.0, 1.0]],
            2.0,
        ),
        (
            [],
            [("Person", "George Washington"), ("Place", "Valencia")],
            False,
            0,
            2,
            [[1.0, 1.0], [1.0, 1.0]],
            2.0,
        ),
    ],
)
def test_oi_ecer_ewer_eval(
    labels,
    predictions,
    char_level,
    expected_num_ne_gt,
    expected_num_ne_hyp,
    expected_cost_matrix,
    expected_errors,
):
    score = OiEcerEwer(labels, predictions, char_level)
    assert score.num_ne_gt == expected_num_ne_gt
    assert score.num_ne_hyp == expected_num_ne_hyp
    assert score.costs == expected_cost_matrix
    assert score.errors == expected_errors


@pytest.mark.parametrize(
    (
        "labels",
        "predictions",
        "nerval_threshold",
        "expected_num_ne_gt",
        "expected_num_ne_hyp",
        "expected_cost_matrix",
        "expected_true_positives",
        "expected_false_positives",
        "expected_false_negatives",
    ),
    [
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Person", "George Washingtn")],
            0.0,
            2,
            1,
            [[2.0, 1.0], [2.0, 1.0]],
            0,
            1,
            2,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Person", "George Washingtn")],
            30.0,
            2,
            1,
            [[0.0, 1.0], [2.0, 1.0]],
            1,
            0,
            1,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Person", "George Washingtn")],
            100.0,
            2,
            1,
            [[0.0, 1.0], [2.0, 1.0]],
            1,
            0,
            1,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Place", "Valencia")],
            0.0,
            2,
            1,
            [[2.0, 1.0], [0.0, 1.0]],
            1,
            0,
            1,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Place", "Valencia")],
            30.0,
            2,
            1,
            [[2.0, 1.0], [0.0, 1.0]],
            1,
            0,
            1,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [("Place", "Valencia")],
            100.0,
            2,
            1,
            [[2.0, 1.0], [0.0, 1.0]],
            1,
            0,
            1,
        ),
        (
            [("Person", "George Washington"), ("Place", "Valencia")],
            [],
            0.0,
            2,
            0,
            [[1.0, 1.0], [1.0, 1.0]],
            0,
            0,
            2,
        ),
        (
            [],
            [("Person", "George Washington"), ("Place", "Valencia")],
            0.0,
            0,
            2,
            [[1.0, 1.0], [1.0, 1.0]],
            0,
            2,
            0,
        ),
    ],
)
def test_oi_nerval_eval(
    labels,
    predictions,
    nerval_threshold,
    expected_num_ne_gt,
    expected_num_ne_hyp,
    expected_cost_matrix,
    expected_true_positives,
    expected_false_positives,
    expected_false_negatives,
):
    score = OiNerval(labels, predictions, nerval_threshold)
    assert expected_num_ne_gt == score.num_ne_gt
    assert expected_num_ne_hyp == score.num_ne_hyp
    assert expected_cost_matrix == score.costs
    assert expected_true_positives == score.true_positives
    assert expected_false_positives == score.false_positives
    assert expected_false_negatives == score.false_negatives


@pytest.mark.parametrize(
    (
        "label_dir",
        "prediction_dir",
        "shuffled_prediction_dir",
        "expected_table",
    ),
    [
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            """| Category | ECER (%) | EWER (%) | N entities | N documents |
|:---------|:--------:|:--------:|-----------:|------------:|
| total    |  21.33   |  28.27   |         28 |           5 |""",
        ),
    ],
)
def test_compute_oi_ecer_ewer(
    label_dir,
    prediction_dir,
    shuffled_prediction_dir,
    expected_table,
):
    table = oiecerewer(label_dir, prediction_dir)
    table_shuffled = oiecerewer(
        label_dir,
        shuffled_prediction_dir,
    )
    assert table.get_string(sortby="Category") == expected_table
    assert table_shuffled.get_string(sortby="Category") == expected_table


@pytest.mark.parametrize(
    (
        "label_dir",
        "prediction_dir",
        "shuffled_prediction_dir",
        "expected_table",
    ),
    [
        (
            FIXTURES / "labels",
            FIXTURES / "predictions",
            FIXTURES / "shuffled_predictions",
            """| Category | Precision (%) | Recall (%) | F1 (%) | N entities | N documents |
|:---------|:-------------:|:----------:|:------:|-----------:|------------:|
| total    |     82.14     |   82.14    | 82.14  |         28 |           5 |""",
        ),
    ],
)
def test_compute_nerval(
    label_dir,
    prediction_dir,
    shuffled_prediction_dir,
    expected_table,
):
    table = oinerval(label_dir, prediction_dir, nerval_threshold=30)
    table_shuffled = oinerval(
        label_dir,
        shuffled_prediction_dir,
        nerval_threshold=30.0,
    )
    assert table.get_string(sortby="Category") == expected_table
    assert table_shuffled.get_string(sortby="Category") == expected_table



import re
from pathlib import Path

from setuptools import find_packages, setup

SUBMODULE_PATTERN = re.compile("-e ((?:(?!#egg=).)*)(?:#egg=)?(.*)")


def parse_requirements_line(line: str) -> str:
    # Special case for git requirements
    if line.startswith("git+http"):
        assert "@" in line, "Branch should be specified with suffix (ex: @master)"
        assert "#egg=" in line, (
            "Package name should be specified with suffix (ex: #egg=kraken)"
        )
        package_name: str = line.split("#egg=")[-1]
        return f"{package_name} @ {line}"
    # Special case for submodule requirements
    elif line.startswith("-e"):
        package_path, package_name = SUBMODULE_PATTERN.match(line).groups()
        package_path: Path = Path(package_path).resolve()
        # Package name is optional: use folder name by default
        return f"{package_name or package_path.name} @ file://{package_path}"
    else:
        return line


def parse_requirements(filename: str) -> list[str]:
    path = Path(__file__).parent.resolve() / filename
    assert path.exists(), f"Missing requirements: {path}"
    return list(
        map(parse_requirements_line, map(str.strip, path.read_text().splitlines())),
    )


setup(
    install_requires=parse_requirements("requirements.txt"),
    extras_require={
        "dev": ["pre-commit", "tox"],
    },
    packages=find_packages(),
)



"""Compute the ECER/EWER/Nerval metrics from a label/prediction dataset."""

import logging

from prettytable import PrettyTable

from ie_eval.metrics.cer import cerwer, oiecerewer
from ie_eval.metrics.nerval import nerval, oinerval

logger = logging.getLogger(__name__)


def compute_ecerewer(
    order_independent: bool = False,
    print_table: bool = True,
    *args,
    **kwargs,
) -> PrettyTable:
    """Compute reading-order-independent ECER and EWER metrics.

    Args:
        print_table (bool, optional): Whether to print the table. Defaults to True.
        order_independent (bool): Whether the metric should be order-independent.

    Returns:
        PrettyTable: The evaluation table formatted in Markdown.
    """
    table = (
        oiecerewer(*args, **kwargs) if order_independent else cerwer(*args, **kwargs)
    )

    if print_table:
        print(table)  # noqa: T201
    return table


def compute_nerval(
    order_independent: bool = False,
    print_table: bool = True,
    *args,
    **kwargs,
) -> PrettyTable:
    """Read BIO files and compute Precision, Recall and F1 globally or for each NER category.

    Args:
        print_table (bool, optional): Whether to print the table. Defaults to True.
        order_independent (bool): Whether the metric should be order-independent.

    Returns:
        A Markdown formatted table containing evaluation results.
    """
    table = oinerval(*args, **kwargs) if order_independent else nerval(*args, **kwargs)

    if print_table:
        print(table)  # noqa: T201
    return table



"""Scorer."""

from collections import Counter, defaultdict
from dataclasses import dataclass, field
from typing import NamedTuple

import numpy as np
from munkres import Munkres

# Global Munkres algorithm
M = Munkres()


# Global function for ECER, EWER and NERVAL
def calc_dist_sus_entity(
    hyp_ne: tuple[str, str],
    gt_ne: tuple[str, str],
    char_level: bool,
) -> float:
    """Calculate substitution distance between 2 entities (hyp_ne, gt_ne).

    Args:
        hyp_ne (tuple): hypothesized Named Entity, format: (category, transcription)
        gt_ne (tuple): label Named Entity, format: (category, transcription)
        char_level (bool): if True, evaluate at character level.

    Returns:
        float: edit distance in range [0.0, 1.0]
    """
    # Check coincidence of NE category
    if hyp_ne[0] != gt_ne[0]:
        return 1.0

    hyp_word_transcription = hyp_ne[1]
    gt_word_transcription = gt_ne[1]

    if char_level is False:
        # Split by word
        hyp_word_transcription = hyp_word_transcription.split()
        gt_word_transcription = gt_word_transcription.split()

    # Tuples of (distance, correct tokens )
    vec_dist_pre = [(i, 0) for i in range(len(gt_word_transcription) + 1)]
    vec_dist_act = [(0, 0)] * (len(gt_word_transcription) + 1)

    # if char_level == true, then the string is explored character by character (including space)
    for j in range(len(hyp_word_transcription)):
        vec_dist_act[0] = (j + 1, 0)
        for i in range(len(gt_word_transcription)):
            dist_ins = (vec_dist_act[i][0] + 1, vec_dist_act[i][1])
            dist_bor = (vec_dist_pre[i + 1][0] + 1, vec_dist_pre[i + 1][1])

            cost_sus = int(hyp_word_transcription[j] != gt_word_transcription[i])
            dist_sus = (
                vec_dist_pre[i][0] + cost_sus,
                vec_dist_pre[i][1] + (1 - cost_sus),
            )

            vec_dist_act[i + 1] = min(dist_ins, dist_bor, dist_sus)

        vec_dist_pre, vec_dist_act = vec_dist_act, vec_dist_pre

    # Saturation of CER/WER (min(CER, 1.0))
    return min(float(vec_dist_pre[-1][0]) / float(len(gt_word_transcription)), 1.0)


@dataclass
class OiEcerEwer:
    """Base class for order independent ECER / EWER computation."""

    labels: list[tuple[str, str]]
    predictions: list[tuple[str, str]]
    compute_ecer: bool

    costs: list[list[float]] = field(default_factory=list)
    errors: float = 0.0

    def __post_init__(self):
        """After object creation, generate cost matrix and compute errors."""
        # Generate cost matrix
        self._generate_cost_matrix()

        # Generate error
        self._compute_errors()

    @property
    def num_ne_gt(self) -> int:
        """Compute number of NEs in the label."""
        return len(self.labels)

    @property
    def num_ne_hyp(self) -> int:
        """Compute number of NEs in the prediction."""
        return len(self.predictions)

    def _generate_cost_matrix(self) -> None:
        """Generation of square cost matrix."""
        max_num_ne = max(self.num_ne_gt, self.num_ne_hyp)
        for i in range(max_num_ne):
            self.costs.append([1.0] * max_num_ne)

        # Population of cost matrix
        for i in range(max_num_ne):
            gt_ne_i = False if i >= self.num_ne_gt else self.labels[i]
            for j in range(max_num_ne):
                hyp_ne_j = False if j >= self.num_ne_hyp else self.predictions[j]
                if not gt_ne_i or not hyp_ne_j:
                    # One of them does not exist => Padding with dummy symbols
                    # Cost equal to insertion or deletion of GT or HYP NE, respectively
                    self.costs[i][j] = 1.0
                else:
                    # Both of them exist: cost equal to substitution
                    self.costs[i][j] = calc_dist_sus_entity(
                        hyp_ne_j,
                        gt_ne_i,
                        char_level=self.compute_ecer,
                    )

    def _compute_errors(self) -> None:
        """Compute the best entity assignment between labels and predictions at character level."""
        self.errors = 0.0

        if self.num_ne_hyp == 0 and self.num_ne_gt == 0:
            # No errors
            self.errors = 0.0
            # TODO: https://gitlab.teklia.com/ner/metrics/ie-eval/-/issues/7

        elif (self.num_ne_hyp != 0 and self.num_ne_gt == 0) or (
            self.num_ne_hyp == 0 and self.num_ne_gt != 0
        ):
            # Only one exists (XOR) -> All detected NEs are errors
            self.errors = self.num_ne_hyp + self.num_ne_gt
            # TODO: https://gitlab.teklia.com/ner/metrics/ie-eval/-/issues/7

        else:
            # General case computation
            # Execute Munkres/Hungarian algorithm
            indexes = M.compute(self.costs)
            for i, j in indexes:
                self.errors += self.costs[i][j]


@dataclass
class OiNerval:
    """Base class for order independent Nerval computation of Precision, Recall and F1 scores."""

    labels: list[tuple[str, str]]
    predictions: list[tuple[str, str]]
    nerval_threshold: float = 0.0

    costs: list[list[float]] = field(default_factory=list)

    true_positives: int = 0
    false_positives: int = 0
    false_negatives: int = 0

    def __post_init__(self):
        """After object creation, generate cost matrix and compute errors."""
        # Generate cost matrix
        self._generate_cost_matrix()

        # Generate error
        self._compute_metric()

    @property
    def num_ne_gt(self) -> int:
        """Returns the number of NEs in the label."""
        return len(self.labels)

    @property
    def num_ne_hyp(self) -> int:
        """Returns the number of NEs in the prediction."""
        return len(self.predictions)

    def _generate_cost_matrix(self) -> None:
        """Generation of square cost matrix."""
        fixed_nerval_threshold = self.nerval_threshold / 100.0
        max_num_ne = max(self.num_ne_gt, self.num_ne_hyp)

        for i in range(max_num_ne):
            self.costs.append([1.0] * max_num_ne)

        # Population of cost matrix
        for i in range(max_num_ne):
            gt_ne_i = False if i >= self.num_ne_gt else self.labels[i]
            for j in range(max_num_ne):
                hyp_ne_j = False if j >= self.num_ne_hyp else self.predictions[j]
                if not gt_ne_i or not hyp_ne_j:
                    # One of them does not exist => Padding with dummy symbols
                    # Cost equal to insertion or deletion of GT or HYP NE, respectively
                    self.costs[i][j] = 1.0
                else:
                    # Both of them exist: cost equal to substitution
                    self.costs[i][j] = 2 * calc_dist_sus_entity(
                        hyp_ne_j,
                        gt_ne_i,
                        char_level=True,
                    )
                    if self.costs[i][j] > 0 and gt_ne_i[0] == hyp_ne_j[0]:
                        # If there are character errors but tags match
                        # Set to extremes
                        self.costs[i][j] = (
                            0 if self.costs[i][j] <= (2 * fixed_nerval_threshold) else 2
                        )

    def _compute_metric(self) -> None:
        """Compute best entity assignment and return True Positives, False Positives, False Negatives."""
        if self.num_ne_gt == 0 and self.num_ne_hyp == 0:
            # No errors, TP = FP = FN = 0
            self.true_positives = 0
            self.false_positives = 0
            self.false_negatives = 0
            return

        if self.num_ne_hyp != 0 and self.num_ne_gt == 0:
            # Only hypothesis exists, all NEs are FP
            self.true_positives = 0
            self.false_positives = self.num_ne_hyp
            self.false_negatives = 0
            return

        if self.num_ne_gt != 0 and self.num_ne_hyp == 0:
            # Only GT exists, all NEs are FN
            self.true_positives = 0
            self.false_positives = 0
            self.false_negatives = self.num_ne_gt
            return

        # General case computation
        indexes = M.compute(self.costs)

        tp = 0
        fp = 0
        fn = 0

        for i, j in indexes:
            # Depending on match cost, decide
            if self.costs[i][j] == 0:  # Substitution without cost, TP
                tp = tp + 1
            elif self.costs[i][j] == 2:  # Substitution with cost = 2, FP and FN
                fn = fn + 1
                fp = fp + 1
            elif (
                self.costs[i][j] == 1
            ):  # Insertion or deletion with cost = 1, may be FP or FN
                if (
                    i >= self.num_ne_gt
                ):  # GT NE does not exist, therefore hypothesized NE is FP
                    fp = fp + 1
                elif (
                    j >= self.num_ne_hyp
                ):  # Hyp NE does not exist, therefore GT NE is FN
                    fn = fn + 1
                else:
                    raise Exception(
                        f"Error in substitution between NEs GT: {i} - Hyp: {j}. Cost is 1 but both exist",
                    )
            else:
                raise Exception(
                    "Cost matrix was not properly initialized in Nerval computation",
                )

        self.true_positives = tp
        self.false_positives = fp
        self.false_negatives = fn


class BagOfWords(NamedTuple):
    """Base class for bag-of-word metrics. Extension of bWER defined in End-to-End Page-Level Assessment of Handwritten Text Recognition (https://arxiv.org/pdf/2301.05935.pdf)."""

    labels: list[str | tuple[str, str]]
    predictions: list[str | tuple[str, str]]

    @property
    def label_counter(self) -> Counter[str]:
        """Split the label into a list of words."""
        return Counter(self.labels)

    @property
    def prediction_counter(self) -> Counter[str]:
        """Split the prediction into a list of words."""
        return Counter(self.predictions)

    @property
    def true_positives(self) -> int:
        """Count true positive words."""
        return sum((self.label_counter & self.prediction_counter).values())

    @property
    def false_positives(self) -> int:
        """Count false positive words."""
        return sum((self.prediction_counter - self.label_counter).values())

    @property
    def false_negatives(self) -> int:
        """Count false negatives words."""
        return sum((self.label_counter - self.prediction_counter).values())

    @property
    def all_words(self) -> list[str | tuple[str, str]]:
        """All tagged words."""
        return sorted(set(self.labels + self.predictions))

    @property
    def label_word_vector(self) -> np.array:
        """Iterate over the set of tagged words and count occurrences in the label."""
        return np.array(
            [self.labels.count(w) for w in self.all_words],
        )

    @property
    def prediction_word_vector(self) -> np.array:
        """Iterate over the set of words and count occurrences in the prediction."""
        return np.array(
            [self.predictions.count(word) for word in self.all_words],
        )

    @property
    def insertions_deletions(self) -> int:
        """Count unavoidable insertions and deletions. See Equation 8 from https://arxiv.org/pdf/2301.05935.pdf."""
        return abs(len(self.labels) - len(self.predictions))

    @property
    def substitutions(self) -> int:
        """Count substitutions. See Equation 8 from https://arxiv.org/pdf/2301.05935.pdf."""
        return (
            np.absolute(self.prediction_word_vector - self.label_word_vector).sum()
            - self.insertions_deletions
        ) / 2

    @property
    def errors(self) -> int:
        """Count total number of errors."""
        return self.substitutions + self.insertions_deletions


class MicroAverageErrorRate:
    """Compute total error rates."""

    def __init__(self) -> None:
        """Initialize errors and counts.

        Examples:
            >>> score = MicroAverageErrorRate()
        """
        self.label_word_count = defaultdict(int)
        self.error_count = defaultdict(int)
        self.count = defaultdict(int)

    def update(self, key: str, score: BagOfWords | OiEcerEwer) -> None:
        """Update the score with the current evaluation for a given key.

        Args:
            key (str): Category to update.
            score (BagOfWords | OiEcerEwer): Current score.

        Examples:
            >>> score.update("total", [("person", "Georges"), ("person", "Washington")])
        """
        self.label_word_count[key] += len(score.labels)
        self.count[key] += 1
        self.error_count[key] += score.errors

    @property
    def error_rate(self) -> dict[str, float]:
        """Error rate for each key."""
        return {
            key: min(100 * self.error_count[key] / self.label_word_count[key], 100)
            for key in self.label_word_count
        }

    @property
    def categories(self) -> list[str]:
        """Get all categories in the label."""
        return list(self.label_word_count.keys())


class MicroAverageFScore:
    """Compute total precision, recall, and f1 scores."""

    def __init__(self) -> None:
        """Initialize error counts.

        Examples:
            >>> score = MicroAverageFScore()
        """
        self.label_word_count = defaultdict(int)
        self.count = defaultdict(int)
        self.true_positives = defaultdict(int)
        self.false_positives = defaultdict(int)
        self.false_negatives = defaultdict(int)

    def update(self, key: str, score: BagOfWords) -> None:
        """Update the score with the current evaluation for a given key.

        Args:
            key (str): Category to update.
            score (BagOfWords): Current score.

        Examples:
            >>> score.update("total", BagOfWords(label.entities, pred.entities))
        """
        self.label_word_count[key] += len(score.labels)
        self.count[key] += 1
        self.true_positives[key] += score.true_positives
        self.false_positives[key] += score.false_positives
        self.false_negatives[key] += score.false_negatives

    @staticmethod
    def recall_score(true_positives: int, false_negatives: int) -> float:
        """Compute the recall."""
        return (
            100 * true_positives / (true_positives + false_negatives)
            if true_positives + false_negatives > 0
            else 100
        )

    @staticmethod
    def precision_score(true_positives: int, true_negatives: int) -> float:
        """Compute the precision."""
        return (
            100 * true_positives / (true_positives + true_negatives)
            if true_positives + true_negatives > 0
            else 100
        )

    @staticmethod
    def f1_score(precision: float, recall: float) -> float:
        """Compute the F1 score."""
        return (
            2 * precision * recall / (precision + recall)
            if precision + recall > 0
            else 0
        )

    @property
    def recall(self) -> dict[str, float]:
        """Recall score for each key."""
        return {
            key: self.recall_score(self.true_positives[key], self.false_negatives[key])
            for key in self.count
        }

    @property
    def precision(self) -> dict[str, float]:
        """Precision score for each key."""
        return {
            key: self.precision_score(
                self.true_positives[key],
                self.false_positives[key],
            )
            for key in self.count
        }

    @property
    def f1(self) -> dict[str, float]:
        """F1 score for each key."""
        return {
            key: self.f1_score(self.precision[key], self.recall[key])
            for key in self.count
        }

    @property
    def categories(self) -> list[str]:
        """Get all categories in the label."""
        return list(self.label_word_count.keys())
